/*=========================================================================*
 *      File    : majfactor.c
 *
 *      Author  : sis
 *      Created : Fri Jun 10 11:27:29 2005
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: majfactor.c,v 1.2 2005/09/29 20:51:18 pgupta Exp $ */

#include "sis.h"
#include "../pld/pld_int.h"
#include "../simplify/simp_int.h"
#include "th_int.h"
#include <math.h>

majsyn_options_t *opt;
array_t *array_cost;
int recursive;
int majsyn_maxfanin;

static int majfactor_usage ARGS(());
static int majfactor_value ARGS((node_t *, node_t *, char *));
static void majfactor_find_best_div ARGS((divisor_t *, divisor_t *));
static int maj_cost_compare ARGS((int *, int *));

int com_majfactor(network_t **network, int argc, char **argv)
{
  int percent; 

  /* process -p option */
  percent = majfactor_options_init(argc, argv);
  if (percent < 0 || percent > 100 || opt->help) {
    majsyn_options_free(opt);
    return majfactor_usage();
  }

  (void) majfactor_split_network(*network, opt, percent);

  /* free the options at the end */
  majsyn_options_free(opt);
  
  return 1;
}


void majfactor_split_network(network_t *network, majsyn_options_t *opt, int percent)
{
  array_t *order;
  node_t *node;
  int i;
  char *err;
  
  order = network_dfs(network);
  for (i = 0; i < array_n(order); i++) {
    node = array_fetch(node_t *, order, i);

      (void) majfactor_split_node(network, opt, node, percent);
      error_init();
      if (!network_check(network)) {
	err = error_string();
	(void) fprintf(sisout, "%s", err);
      }

  }
  array_free(order);
  (void) network_sweep(network);
}


int majfactor_split_node(network_t *network, majsyn_options_t *opt, node_t *node, int percent)
{
  //  static int majfactor_value();
  static int (*fp_majfactor)() = majfactor_value;
  kern_node *sorted;
  divisor_t *best_div, *div;
  node_t *t_node, *best, *co_best, *rem;
  node_t *p, *q, *r, *pq, *new_node;
  int i, target_index, target_cost;


#ifdef THFACTOR_DEBUG3
  printf("factoring node: ");
  node_print(sisout, node);
  printf("\n");
#endif

  // new termination condition
  if (node_num_fanin(node) <= majsyn_maxfanin) 
    {
      // printf("no need to factor this node.\n");
      return 0;
    }

  /* First try finding kernels */
  if ((t_node = ex_find_divisor_quick(node)) != NIL(node_t)) {
#ifdef THFACTOR_DEBUG0 
    printf("Find a divisor.\n ");
#endif   
    sorted = ALLOC(kern_node, 1);
    sorted->node = node;
    sorted->cost_array = array_alloc(divisor_t *, 0);
    sorted->size = majsyn_maxfanin;

    array_cost = array_alloc(int, 0);
    ex_kernel_gen(node, fp_majfactor, (char *)sorted);
    ex_subkernel_gen(node, fp_majfactor, 0, (char *)sorted);
    // ex_kernel_gen(node, majfactor_value, (char *)sorted);
    // ex_subkernel_gen(node, majfactor_value, 0, (char *)sorted);
    best_div = ALLOC(divisor_t, 1);
    best_div->divisor = NIL(node_t);
    best_div->kernelsize = THFACTOR_HICOST;

    /* sort the 'array_cost' array in ascending order */
#ifdef THFACTOR_DEBUG0 
    printf("\nbefore sorting:\n");
    for(i=0; i< array_n(array_cost); i++)
      printf("%d, ",array_fetch(int, array_cost, i));
#endif

    array_sort(array_cost, maj_cost_compare);
#ifdef THFACTOR_DEBUG1
    printf("\nafter sorting:\n");
    for(i=0; i< array_n(array_cost); i++)
      printf("%d, ",array_fetch(int, array_cost, i));
#endif
    /* find out the x%th cost in ascending order, for comparision basis */
    target_index = array_n(array_cost);
    target_index = (int) ceil(percent * target_index / 100); /* for example, take the first percent% */ 
    target_cost = array_fetch(int, array_cost, target_index - 1);
#ifdef THFACTOR_DEBUG1 
    printf("the %d /100 cost position is %d, cost is %d \n", percent, target_index, target_cost);
#endif

    /* find out the best factorization based on threshold logic */
    for( i=0; i< array_n(sorted->cost_array); i++){
      div = array_fetch(divisor_t *, sorted->cost_array, i);
      if (div->kernelsize <= target_cost)
	(void)majfactor_find_best_div(div, best_div);
    }
    
    array_free(sorted->cost_array);
    FREE(sorted);
    array_free(array_cost);

    
    if(best_div->divisor != NIL(node_t)) {
      best = best_div->divisor;
      FREE(best_div);
      co_best = node_div(node, best, &rem);

#ifdef THFACTOR_DEBUG2 
      (void)fprintf(sisout,"Node %s has been added to the network\n", node_name(best));
      (void)node_print(sisout, best);
      (void)fprintf(sisout, "cokernel");
      node_print(sisout, co_best);
      (void)fprintf(sisout, "remainder");
      node_print(sisout, rem);
      printf("node to be changed:\n");
      node_print(sisout,node);
      printf("node to be substitued:\n");
      node_print(sisout, best);
#endif
      network_add_node(network, best);
      p = node_literal(best, 1);
      q = node_literal(co_best, 1);
      r = node_literal(rem, 1);
      pq = node_and(p, q);
      new_node = node_or(pq, r);
      node_collapse(new_node, co_best);
      node_collapse(new_node, rem);
      node_replace(node, new_node);
      node_free(p);
      node_free(q);
      node_free(pq);
      node_free(r);
      node_free(co_best);
      node_free(rem);

#ifdef THFACTOR_DEBUG2 
      printf("node after change: \n");
      node_print(sisout, node);
#endif

      //it is an option to split a node recursively or not
      if (recursive == 1)
	(void) majfactor_split_node(network, opt, node, percent);
    }
    else {
      ;
      // printf("best_div->divisor == NIL(node_t)\n");
    }
    node_free(t_node);
  }
  else 
    return 0;
  
  return 1;
}


static int majfactor_value(node_t *kernel, node_t *cokernel, char *state)
{
  divisor_t *k_data;
  kern_node *store;
  node_t *rem, *new_cokernel, *tmp_node;
  network_t *tmp_network;
  int cost, size, unate, fanin_count, sop_count, cube_count, node_count;


  cost = 0;
  store = (kern_node *)state;
  size = store->size;
  new_cokernel = node_div(store->node, kernel, &rem);
  
#ifdef THFACTOR_DEBUG1
  printf("....................\n computing kernel values of node:\n");
  node_print(sisout, store->node);
#endif  
  
  ///////////////////////////////////////////////////////////////////////////////////
    cost = node_num_literal(kernel) + node_num_literal(new_cokernel) + node_num_literal(rem);
    k_data = ALLOC(divisor_t, 1);
    k_data->divisor = kernel;
    k_data->kernelsize = cost;
    array_insert_last(divisor_t *, store->cost_array, k_data);
  

    /* also insert the cost for this kernel into a global array */
    array_insert_last(int, array_cost, cost);
#ifdef THFACTOR_DEBUG1
    printf("kernel: %d, cokernel: %d, rem:%d, cost: %d\n",node_num_literal(kernel), node_num_literal(new_cokernel),
	   node_num_literal(rem),cost);
    node_print(sisout, kernel);
    node_print(sisout, new_cokernel);
    node_print(sisout, rem);
#endif

    node_free(rem);
    node_free(cokernel);
}


static void majfactor_find_best_div(divisor_t *div, divisor_t *best_div)
{
  node_t *tmp_node;
  int fanin_count;
  int literal_count, cube_count, node_count, maj_cost, network_area;
  network_t *tmp_network;

  /* scheme 1: call xl_split and MALS on this kernel based new network */
  tmp_node = node_dup(div->divisor);
  //fanin_count = node_num_fanin(tmp_node);
  literal_count = node_num_literal(tmp_node);
  cube_count = node_num_cube(tmp_node);
  tmp_network = network_create_from_node(tmp_node);
  
  if (node_num_fanin(tmp_node) == 1)
    maj_cost =  THFACTOR_HICOST;
  else {
    /* call xl_split to preprocess the network to be a 3-feasible network */
    split_network(tmp_network, 3);    
    /* call MALS on this preprocessed 3-feasible network */
    majsyn_whole_network(&tmp_network, opt);
    node_count = network_num_internal(tmp_network);
    /* the smaller the cost, the better */

    /* cost function 1 */
    //rui TODO: can modify the maj_cost as divided only by literal_count 
  //  maj_cost = node_count * 40 / (literal_count + cube_count);
    maj_cost = node_count * 40 / literal_count;

  }

#ifdef THFACTOR_DEBUG1 
  printf("checking node... for best div:\n");
  node_print(sisout, tmp_node);
  printf("threshold node count = %d, fanin_count = %d, cube_count = %d \n maj_cost = %d \n", 
  	 node_count, fanin_count, cube_count, maj_cost);
#endif
  
  if(maj_cost < best_div->kernelsize) {
    if(best_div->divisor != NIL(node_t )){
      node_free(best_div->divisor);
    }
    best_div->divisor = node_dup(div->divisor);
    best_div->kernelsize = maj_cost;
  } 
  
#ifdef THFACTOR_DEBUG0 
  (void) printf("best_div kernel size: %d, for the node:\n", maj_cost); 
  node_print(sisout, div->divisor);
#endif
  
  node_free(div->divisor);
  network_free(tmp_network);

  FREE(div);
}

/*********************************************************************** 
 * This function itself is used as a parameter in function: array_sort *
 * It compares two minterms by the minterm index                       *
 * returns -1 if n1 is greater, 0 otherwise                            *
 ***********************************************************************/
static int maj_cost_compare(int *c1, int *c2)
{
  /* return appropriate status */
  if (*c1 < *c2)
    return -1;
  else if (*c1 == *c2)
    return 0;
  else
    return 1;
}



/******************************************* 
 * initialize thsyn options data structure *
 *******************************************/
int majfactor_options_init(int argc, char **argv)
{
  int c, error, percent;
  
  opt = ALLOC(majsyn_options_t, 1);

  /* default values */
  majsyn_maxfanin = 3;
  percent = 60;
  recursive = 0;
  opt->m = -1; /* collapse fanin restriction (default no collapsing) */
  opt->help = FALSE;

  /* parse options */
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "n:p:rh")) != EOF) {
    switch (c) {
    case 'n': opt->m = atoi(util_optarg); break;
    case 'h': opt->help = TRUE; break;
    case 'p': percent = atoi(util_optarg); break;
    case 'r': recursive = 1; break;
    default: opt->help = TRUE; break;
    }
  }
  
  return percent;
}


/*******************
 * majfactor usage *
 *******************/
static int majfactor_usage(void)
{
  fprintf(siserr, "usage: majfactor [-n num1] [-p num2] [-r] [-h]\n"); 
  fprintf(siserr, "   Decomposing the nodes for a better majority network.\n");
  fprintf(siserr, "   -n num1:\tAssign the integer num1 to fanin restriction, default value is 3.\n");  
  fprintf(siserr, "   -p num2:\tConsider the num2/100 portion of the candidate kernels (0<= num2 <=100, default is 60).\n");
  fprintf(siserr, "   -r :\tSplit each node in the network recursively until num1 is reached or no kernels (default is recursively.\n");
  fprintf(siserr, "   -h:        \tFor help information. \n");  
  return TH_FAIL;
}
