/*=========================================================================*
 *	File    :	thdot.c
 *
 *	Author  :	Rui Zhang
 *	Created :	Thu Sep 25 23:21:02 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thdot.c,v 1.6 2004/01/16 04:07:39 pgupta Exp $ */


#include "sis.h"
#include "th_int.h"

static int  thdot_usage ARGS(());
static void thdot_dump ARGS((FILE *, network_t *, array_t *, thdot_options_t *));
static void thdot_graph_print ARGS((FILE *, network_t *, thdot_options_t *));
static void thdot_node_print ARGS((FILE *, node_t *, int, thdot_options_t *));
static void thdot_edge_print ARGS((FILE *, array_t *node_vec, node_t *, int, thdot_options_t *));
static array_t *thdot_find_fanin ARGS((node_t *, array_t *, int));
static void thdot_single_edge ARGS((FILE *, int, int, int, thdot_options_t *));

int com_thdot(network_t **network, int argc, char **argv)
{
  FILE *opf;
  array_t *node_vec;
  thdot_options_t *opt;

  /* options parse and checks */
  if ((opt = thdot_options_init(argc, argv)) == NIL(thdot_options_t)
      || opt->help) {
    thdot_options_free(opt);
    return thdot_usage();
  }  
  
  /* check if threshold network */
  if (is_threshold_network(*network) != TH_SUCCESS) {
    thdot_options_free(opt);
    return TH_FAIL;
  }
  
  if ((opf = com_open_file(opt->filename, "w", 0, 0)) == NIL(FILE)) {
    fprintf(siserr, "Unable to open file '%s' to write\n", opt->filename);	
    thdot_options_free(opt);
    return TH_FAIL;
  }

  node_vec = network_dfs(*network);
  //  node_vec = com_get_nodes(*network, argc-util_optind+1, argv+util_optind-1);
  if (array_n(node_vec) < 1)
    return thdot_usage();
  thdot_dump(opf, *network, node_vec, opt);
  /*  don't forget to free space */
  array_free(node_vec);
  fclose(opf);

  /* free options */
  thdot_options_free(opt);

  return TH_SUCCESS;
}


/* dump vcg formats into the file *fp */
static void thdot_dump(FILE *fp, network_t *network, array_t *node_vec, thdot_options_t *opt)
{ 
  int i; 
  node_t *n;

  /* printing graph description */
  thdot_graph_print(fp,network,opt);

  /* printing node and edge descriptions */
  for (i = 0; i < array_n(node_vec); i++) {
    n = array_fetch(node_t *, node_vec,i);
    if (node_type(n) != PRIMARY_OUTPUT) {
      /* printing node description */
      thdot_node_print(fp, n, i, opt);
      /* printing edge description */
      thdot_edge_print(fp, node_vec, n, i, opt);
    }
  }
  /* print end of graph */
  fprintf(fp,"       }\n");
}

static void thdot_graph_print(FILE *fp, network_t *network, thdot_options_t *opt)
{
  /* print graph attributes */  
  fprintf(fp,"digraph \"%s\" {\n", network->net_name);
  fprintf(fp,"         rankdir=LR;\n");
  fprintf(fp,"         size = \"8,5\"; ratio = fill;\n");
  /* print node attributes */
  fprintf(fp,"         node [style=filled, fillcolor=lightyellow, shape = ellipse, color = black, fontcolor =red];\n");
  /* print edge attributes */
  fprintf(fp,"         edge [color = black, fontcolor = red];\n");
}

static void thdot_node_print(FILE *fp, node_t *n, int index, thdot_options_t *opt)
{ 
  int i;
  node_t *fanin;
  node_type_t type;
  thnode_t *th;

  type = node_type(n);
  //  fprintf(fp,"node: { title: \"%d\" ", index);
  if (type == INTERNAL) { 
    if ((type == PRIMARY_INPUT) || (type == PRIMARY_OUTPUT))
      fprintf(fp,"%d [label = \"%s\"];\n", index, node_name(n));
    else {	  
      fprintf(fp,"%d [label = \"", index);
      fprintf(fp, "%s", node_name(n));
      th = (thnode_t *) n->th;
      if (opt->mode == 1) // printing node according to mode 1 format 
	{
	  fprintf(fp, "\\n");
	  fprintf(fp,"%d\"];\n",(int)th->T);
	}
      else // printing node according to mode 2 format
	{
	  fprintf(fp, "; ");
	  foreach_fanin(n, i, fanin)
	    fprintf(fp, "%s,", node_name(fanin));
	  fprintf(fp, "\\n");
	  fprintf(fp,"%d;",(int)th->T);
	  foreach_fanin(n, i, fanin)
	    fprintf(fp, "%d,", (int)th->W[i]);
	  fprintf(fp,"\"];\n");
	}
    }
  }
  else 
    fprintf(fp,"%d [label = \"%s\"];\n", index, node_name(n));
  
  //  fprintf(fp,"}\n");
}

static void thdot_edge_print(FILE *fp, array_t *node_vec, node_t *n, int index, thdot_options_t *opt)
{
 int i, j, w, k;
 node_t *fanin, *source_n, *n_copy, *nin;
 node_type_t type;
 array_t *fi_vec;
 thnode_t *th;

 n_copy = node_dup(n);
 type = node_type(n);	      
 th = (thnode_t *) n->th;

 if (type != PRIMARY_INPUT) {
   if (type == PRIMARY_OUTPUT) {
     fi_vec = thdot_find_fanin(n, node_vec, index);
     if (array_n(fi_vec) > 0)
       for(j = 0; j < array_n(fi_vec); j++)
	 thdot_single_edge(fp, array_fetch(int, fi_vec, j), index, 0, opt);
   }
   else {
     foreach_fanin(n, i, fanin) {
       fi_vec = thdot_find_fanin(fanin, node_vec, index);
       if (array_n(fi_vec) > 0)
	 for(j = 0; j < array_n(fi_vec); j++) {
	   // find out the weight
	   // fetch out the source node
	   source_n = array_fetch(node_t *, node_vec, array_fetch(int, fi_vec, j));
	   foreach_fanin(n_copy, k, nin) {	
	     // find out the index of the weight
	     if(strcmp(node_name(source_n), node_name(nin)) == 0)
	       // get the corresponding weight
	       w = (int)th->W[k]; 
	   }
	   thdot_single_edge(fp, array_fetch(int, fi_vec, j), index, w, opt);
	 }
     }
   }
 }
 FREE(n_copy);
}

/* find all the indexes of the nodes that are the fanout of the given node name */
static array_t *thdot_find_fanin(node_t *des_n, array_t *node_vec, int index)
{
  int i;
  node_t *tmp_n;
  array_t *fi_nodes;  
  
  /* allocate the array to save the fanin nodes' index */
  fi_nodes = array_alloc(int,0);
  for (i = 0; i < array_n(node_vec); i++)
    if ( i != index)  {  
      tmp_n = array_fetch(node_t *,node_vec,i);
      /* find a fanout node */
      if(strcmp(node_name(des_n), node_name(tmp_n)) == 0) {
	// save this index i into a array_t *fi_nodes
	array_insert_last(int, fi_nodes, i);
      }
    }
  return fi_nodes;
}

/* given the source and destination of each node, print out the 
 * corresponding single edge
 */
static void thdot_single_edge(FILE *fp, int source, int desti, int weight, thdot_options_t *opt)
{
  if (opt->mode == 1) // printing node according to mode 1 format 
    fprintf(fp,"%d -> %d [label=\"%d\"];\n", source, desti, weight);
  else
    fprintf(fp,"%d -> %d;\n", source, desti);
}

/* initialize thdot options data structrure */
thdot_options_t *thdot_options_init(int argc, char **argv)
{
  int c;
  thdot_options_t *opt;
  
   opt = ALLOC(thdot_options_t, 1);
   /* default values */
   opt->mode = 1;
   opt->filename = util_strsav("thdot_output.dot");
   opt->help = FALSE;

  /* parse options */
  util_getopt_reset();
 while ((c = util_getopt(argc, argv, "o:m:h")) != EOF) {
    switch (c) {
    case 'o':
      opt->filename = util_strsav(util_optarg);
      break;
    case 'm':
      opt->mode = atoi(util_optarg);
      if (!((opt->mode == 1) || (opt->mode == 2)))
	opt->help = TRUE;
      break;
    case 'h':
      opt->help = TRUE; break;
    default:
      opt->help = TRUE; break;
    }
 }
 return opt;
}


/* free options data structure */
void thdot_options_free(thdot_options_t *opt)
{
  if (opt == NIL(thdot_options_t))
    return;
  if (opt->filename != NIL(char))
    FREE(opt->filename);
  FREE(opt);
}


/* thdot usage */
static int thdot_usage(void)
{
  // do the usage information here
  fprintf(siserr, "usage: thdot [-o file] [-m int1] [-h]\n");
  fprintf(siserr, "    -o <file> \tWrite the dot file to <file>, default write to thdot_output.dot.\n");
  fprintf(siserr, "    -m <int1> \tAssign the printing mode to int1. Print weight on the edge when int1=1.\n");
  fprintf(siserr, "              \tPrint weight-threshold vector inside of the node when int1=2\n");
  fprintf(siserr, "              \t(default is 1).\n");
  fprintf(siserr, "    -h        \tPrint out help information\n");
  return TH_FAIL;
}
