/*=========================================================================*
 *      File    : thfactor.c
 *
 *      Author  : sis
 *      Created : Thu Apr 21 13:14:35 2005
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thfactor.c,v 1.2 2005/09/29 20:51:18 pgupta Exp $ */

#include "sis.h"
#include "../pld/pld_int.h"
#include "th_int.h"
#include <math.h>

thsyn_options_t *opt;
array_t *array_cost;
int recursive;

static int thfactor_usage ARGS(());
static int thfactor_value ARGS((node_t *, node_t *, char *));
static void thfactor_find_best_div ARGS((divisor_t *, divisor_t *));
static int cost_compare ARGS((int *, int *));

int com_thfactor(network_t **network, int argc, char **argv)
{
  int percent; 

  /* process -p option */
  percent = thfactor_options_init(argc, argv);
  if (percent < 0 || percent > 100 || opt->help) {
    thsyn_options_free(opt);
    return thfactor_usage();
  }

  (void) thfactor_split_network(*network, opt, percent);

  /* free the options at the end */
  thsyn_options_free(opt);
  
  return 1;
}


void thfactor_split_network(network_t *network, thsyn_options_t *opt, int percent)
{
  array_t *order;
  node_t *node;
  int i;
  char *err;
  
  order = network_dfs(network);
  for (i = 0; i < array_n(order); i++) {
    node = array_fetch(node_t *, order, i);

      (void) thfactor_split_node(network, opt, node, percent);
      error_init();
      if (!network_check(network)) {
	err = error_string();
	(void) fprintf(sisout, "%s", err);
      }

  }
  array_free(order);
  (void) network_sweep(network);
}


int thfactor_split_node(network_t *network, thsyn_options_t *opt, node_t *node, int percent)
{
  static int (*fp_thfactor)() = thfactor_value;
  //  static int thfactor_value();
  kern_node *sorted;
  divisor_t *best_div, *div;
  node_t *t_node, *best, *co_best, *rem;
  node_t *p, *q, *r, *pq, *new_node;
  int i, target_index, target_cost;


#ifdef THFACTOR_DEBUG3
  printf("factoring node: ");
  node_print(sisout, node);
  printf("\n");
#endif

  // new termination condition
  if (node_num_fanin(node) <= opt->max_fanin) 
    {
      // printf("no need to factor this node.\n");
      return 0;
    }

  /* First try finding kernels */
  if ((t_node = ex_find_divisor_quick(node)) != NIL(node_t)) {
#ifdef THFACTOR_DEBUG0 
    printf("Find a divisor.\n ");
#endif   
    sorted = ALLOC(kern_node, 1);
    sorted->node = node;
    sorted->cost_array = array_alloc(divisor_t *, 0);
    sorted->size = opt->max_fanin;

    array_cost = array_alloc(int, 0);
    ex_kernel_gen(node, fp_thfactor, (char *)sorted);
    ex_subkernel_gen(node, fp_thfactor, 0, (char *)sorted);
    //    ex_kernel_gen(node, thfactor_value, (char *)sorted);
    //    ex_subkernel_gen(node, thfactor_value, 0, (char *)sorted);
    best_div = ALLOC(divisor_t, 1);
    best_div->divisor = NIL(node_t);
    best_div->kernelsize = THFACTOR_HICOST;

    /* sort the 'array_cost' array in ascending order */
#ifdef THFACTOR_DEBUG0 
    printf("\nbefore sorting:\n");
    for(i=0; i< array_n(array_cost); i++)
      printf("%d, ",array_fetch(int, array_cost, i));
#endif

    array_sort(array_cost, cost_compare);
#ifdef THFACTOR_DEBUG1
    printf("\nafter sorting:\n");
    for(i=0; i< array_n(array_cost); i++)
      printf("%d, ",array_fetch(int, array_cost, i));
#endif
    /* find out the x%th cost in ascending order, for comparision basis */
    target_index = array_n(array_cost);
    target_index = (int) ceil(percent * target_index / 100); /* for example, take the first percent% */ 
    target_cost = array_fetch(int, array_cost, target_index - 1);
#ifdef THFACTOR_DEBUG1 
    printf("the %d /100 cost position is %d, cost is %d \n", percent, target_index, target_cost);
#endif

    /* find out the best factorization based on threshold logic */
    for( i=0; i< array_n(sorted->cost_array); i++){
      div = array_fetch(divisor_t *, sorted->cost_array, i);
      if (div->kernelsize <= target_cost)
	(void)thfactor_find_best_div(div, best_div);
    }
    
    array_free(sorted->cost_array);
    FREE(sorted);
    array_free(array_cost);

    
    if(best_div->divisor != NIL(node_t)) {
      best = best_div->divisor;
      FREE(best_div);
      co_best = node_div(node, best, &rem);

#ifdef THFACTOR_DEBUG2 
      (void)fprintf(sisout,"Node %s has been added to the network\n", node_name(best));
      (void)node_print(sisout, best);
      (void)fprintf(sisout, "cokernel");
      node_print(sisout, co_best);
      (void)fprintf(sisout, "remainder");
      node_print(sisout, rem);
      printf("node to be changed:\n");
      node_print(sisout,node);
      printf("node to be substitued:\n");
      node_print(sisout, best);
#endif
      network_add_node(network, best);
      p = node_literal(best, 1);
      q = node_literal(co_best, 1);
      r = node_literal(rem, 1);
      pq = node_and(p, q);
      new_node = node_or(pq, r);
      node_collapse(new_node, co_best);
      node_collapse(new_node, rem);
      node_replace(node, new_node);
      node_free(p);
      node_free(q);
      node_free(pq);
      node_free(r);
      node_free(co_best);
      node_free(rem);

#ifdef THFACTOR_DEBUG2 
      printf("node after change: \n");
      node_print(sisout, node);
#endif

      // it is an option to split a node recursively or not
      if (recursive == 1)
	(void) thfactor_split_node(network, opt, node, percent);
    }
    else {
      ;
      // printf("best_div->divisor == NIL(node_t)\n");
    }
    node_free(t_node);
  }
  else 
    return 0;
  
  return 1;
}


static int thfactor_value(node_t *kernel, node_t *cokernel, char *state)
{
  divisor_t *k_data;
  kern_node *store;
  node_t *rem, *new_cokernel, *tmp_node;
  network_t *tmp_network;
  int cost, size, unate, fanin_count, sop_count, cube_count, node_count;


  cost = 0;
  store = (kern_node *)state;
  size = store->size;
  new_cokernel = node_div(store->node, kernel, &rem);
  
#ifdef THFACTOR_DEBUG1
  printf("....................\n computing kernel values of node:\n");
  node_print(sisout, store->node);
#endif  
  
  ///////////////////////////////////////////////////////////////////////////////////
    cost = node_num_literal(kernel) + node_num_literal(new_cokernel) + node_num_literal(rem);
    k_data = ALLOC(divisor_t, 1);
    k_data->divisor = kernel;
    k_data->kernelsize = cost;
    array_insert_last(divisor_t *, store->cost_array, k_data);
  

    /* also insert the cost for this kernel into a global array */
    array_insert_last(int, array_cost, cost);
#ifdef THFACTOR_DEBUG1
    printf("kernel: %d, cokernel: %d, rem:%d, cost: %d\n",node_num_literal(kernel), node_num_literal(new_cokernel),
	   node_num_literal(rem),cost);
    node_print(sisout, kernel);
    node_print(sisout, new_cokernel);
    node_print(sisout, rem);
#endif

    node_free(rem);
    node_free(cokernel);
  ////////////////////////////////////////////////////////////////////////////////////////////
  //  /* compute the kernel value for this factorization */
  //  /* can try different schemes here */
  //
  //  /* scheme 1: call TELS on this kernel based new network */
  //  tmp_node = node_dup(kernel);
  //  fanin_count = node_num_fanin(tmp_node);
  //  cube_count = node_num_cube(tmp_node);
  //  tmp_network = network_create_from_node(tmp_node);
  //
  //  /* call TELS on this new network */
  //  thsyn_whole_network(&tmp_network, opt);
  //  node_count = network_num_internal(tmp_network);
  //
  //  /* the smaller the cost, the better */
  //  cost = node_count * 40 / (fanin_count + cube_count); 
  //
  //#ifdef THFACTOR_DEBUG0 
  //  node_print(sisout, tmp_node);
  //  printf("threshold count = %d, fanin_count = %d, cube_count = %d \n cost = %d \n", 
  //	 node_count, fanin_count, cube_count, cost);
  //#endif
  //
  //  k_data = ALLOC(divisor_t, 1);
  //  k_data->divisor = kernel;
  //  k_data->kernelsize = cost;
  //  array_insert_last(divisor_t *, store->cost_array, k_data);
  //  node_free(rem);
  //  node_free(cokernel);
  //  network_free(tmp_network);
  /////////////////////////////////////////////////////////////////////////////////////////////
}


static void thfactor_find_best_div(divisor_t *div, divisor_t *best_div)
{
  node_t *tmp_node;
  int fanin_count;
  int literal_count, cube_count, node_count, thres_cost, network_area;
  network_t *tmp_network;

  /* scheme 1: call TELS on this kernel based new network */
  tmp_node = node_dup(div->divisor);
  //rui
  //fanin_count = node_num_fanin(tmp_node);
  literal_count = node_num_literal(tmp_node);
  cube_count = node_num_cube(tmp_node);
  tmp_network = network_create_from_node(tmp_node);
  
  if (node_num_fanin(tmp_node) == 1)
    thres_cost =  THFACTOR_HICOST;
  else {
    /* call TELS on this new network */
    thsyn_whole_network(&tmp_network, opt);
    node_count = network_num_internal(tmp_network);
    /* the smaller the cost, the better */

    /* cost function 1 */
    //rui
    //////////thres_cost = node_count * 40 / (fanin_count + cube_count); 
    thres_cost = node_count * 40 / (literal_count + cube_count);

    /* cost function 2: no. of threshold nodes */
    // thres_cost = node_count;

    /* cost function 3: threshold network area */
//	network_area = (int) thsyn_area(tmp_network);
  //      thres_cost = network_area * 40 / (literal_count + cube_count);

    /* cost function 4: threshold network area */
    //    network_area = (int) thsyn_area(tmp_network);
      //  thres_cost = network_area * 40 / (fanin_count + cube_count);

  }

#ifdef THFACTOR_DEBUG1 
  printf("checking node... for best div:\n");
  node_print(sisout, tmp_node);
  printf("threshold node count = %d, fanin_count = %d, cube_count = %d \n thres_cost = %d \n", 
  	 node_count, fanin_count, cube_count, thres_cost);
#endif
  
  if(thres_cost < best_div->kernelsize) {
    if(best_div->divisor != NIL(node_t )){
      node_free(best_div->divisor);
    }
    best_div->divisor = node_dup(div->divisor);
    best_div->kernelsize = thres_cost;
  } 
  
#ifdef THFACTOR_DEBUG0 
  (void) printf("best_div kernel size: %d, for the node:\n", thres_cost); 
  node_print(sisout, div->divisor);
#endif
  
  node_free(div->divisor);
  network_free(tmp_network);

  FREE(div);
}

/*********************************************************************** 
 * This function itself is used as a parameter in function: array_sort *
 * It compares two minterms by the minterm index                       *
 * returns -1 if n1 is greater, 0 otherwise                            *
 ***********************************************************************/
static int cost_compare(int *c1, int *c2)
{
  /* return appropriate status */
  if (*c1 < *c2)
    return -1;
  else if (*c1 == *c2)
    return 0;
  else
    return 1;
}



/******************************************* 
 * initialize thsyn options data structure *
 *******************************************/
int thfactor_options_init(int argc, char **argv)
{
  int c, error, percent;
  
  opt = ALLOC(thsyn_options_t, 1);

  /* default values */
  percent = 60;
  recursive = 0;
  opt->max_fanin = 6; 
  opt->delta_on = 0;
  opt->delta_off = 1;
  opt->filename = util_strsav("thsyn_ampl");
  opt->help = FALSE;

  /* parse options */
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "c:n:f:p:o:rh")) != EOF) {
    switch (c) {
    case 'c': opt->max_fanin = atoi(util_optarg); break;
    case 'n': opt->delta_on = atoi(util_optarg); break;
    case 'f': opt->delta_off = atoi(util_optarg); break;
    case 'p': percent = atoi(util_optarg); break;
    case 'o': opt->filename = util_strsav(util_optarg); break;
    case 'r': recursive = 1; break;
    case 'h': opt->help = TRUE; break;
    default: opt->help = TRUE; break;
    }
  }

  /* check if ampl/cplex are present */
  error = 0;
  if (util_path_search(AMPL) == NIL(char)) {
    fprintf(siserr, "AMPL not found. Please download from www.ampl.com and install\n");
    percent = -1;
    error = 1;
  }
  if (util_path_search(ILP_SOLVER) == NIL(char)) {
    fprintf(siserr, "CPLEX not found. Please download from www.cplex.com and insall\n");
    percent = -1;
    error = 1;
  }
  
  if (opt->max_fanin < 1 || !(opt->delta_on >= (-opt->delta_off)) || error) {
    return -1;
  }
  
  return percent;
}


/******************
 * thfactor usage *
 ******************/
static int thfactor_usage(void)
{
  fprintf(siserr, "usage: thfactor [-c num1] [-n num2] [-f num3] [-p num4] [-r] [-o file] [-h]\n"); 
  fprintf(siserr, "   -c num1:\tAssign the integer num1 to fanin restriction, default value is 6.\n");
  fprintf(siserr, "   -n num2:\tAssign the integer num2 to delta_on, default value is 0.\n");
  fprintf(siserr, "   -f num3:\tAssign the integer num3 to delta_off, default value is 1.\n");
  fprintf(siserr, "   -r :\tSplit each node in the network recursively until num1 is reached or no kernels (default is recursively.\n");

  fprintf(siserr, "   -p num4:\tConsider the num4/100 portion of the candidate kernels (0<= num4 <=100, default is 60).\n");
  fprintf(siserr, "   -o file:   \tWrite the linear program problem into file.mod and file.dat,\n");
  fprintf(siserr, "              \tdefault write to thsyn_ampl.mod and thsyn_ampl.dat.\n");
  fprintf(siserr, "   -h:        \tFor help information. \n");  
  return TH_FAIL;
}
